/*

*/

import renderer.scene.*;
import renderer.models.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;

/**
   Translate the letter {@link P} along the y-axis while
   rotating around the letter's vertical center line (y=0.5).
<p>
   See the animation <a href="../Example_25.gif">Example_25.gif</a>.
<p>
   <a href="../Example_25.gif"><img src="../Example_25.gif" alt="Example_25.gif"></a>
<p>
   For the i'th frame of the animation, the model matrix is
   <pre>{@code
      M =  Ty(0.5) * Ry(18*i) * Ty(-0.5) * Ty(0.05*i)
   }</pre>
   Where {@code Ry} is rotation around the y-axis and {@code Ty} is
   translation in the y-direction.
*/
public class Example_25
{
   public static void main(String[] args)
   {
      // Create the Scene object that we shall render.
      Scene scene = new Scene();
      // Set up a perspective camera.
      final double right = 2.0;
      final double left = -right;
      final double top = 5.0;
      final double bottom = -1.0;
      final double near = 5.0;
      scene.camera.projPerspective(left, right, bottom, top, near);

      // Create a FrameBuffer to render our scene into.
      int width  = 400;
      int height = 600;
      FrameBuffer fb = new FrameBuffer(width, height);

      // Create a set of x and y axes.
      Model axes = new Axes2D(left, right, bottom, top, 8, 12);
      // Color them red.
      ModelShading.setColor(axes, Color.red);
      // Add the axes to the Scene.
      Position axes_p = new Position(axes);
      scene.addPosition(axes_p);
      // Push the axes away from where the camera is.
      axes_p.matrix = Matrix.translate(0, 0, -near);

      // Create a Model of the letter P.
      P modelP = new P();
      // Color the P blue.
      ModelShading.setColor(modelP, Color.blue);
      // Add the model to the Scene.
      Position modelP_p = new Position(modelP);
      scene.addPosition(modelP_p);

      // Translate along the y-axis while rotating around
      // the letter's vertical center line (y=0.5).
      for (int i = 0; i <= 100; i++)
      {
         // Push the model P away from where the camera is.
         modelP_p.matrix = Matrix.translate(0, 0, -near);

         // Rotate P around its center axis.
         modelP_p.matrix.mult( Matrix.translate(0.5, 0, 0) );
         modelP_p.matrix.mult( Matrix.rotateY(i*18) );
         modelP_p.matrix.mult( Matrix.translate(-0.5, 0, 0) );
         // Translate P along the y-axis.
         modelP_p.matrix.mult( Matrix.translate(0, i*0.05, 0) );

         // Render again.
         fb.clearFB(Color.darkGray);
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_Example_25_Frame%03d.ppm", i));
      }
   }
}
